/* Matrox - TVOut class for the Matrox cards
**		Based on code originally developed by Rudolf Cornelissen's vidscreen utility
**
** Copyright (c) 2004 pinc Software. All Rights Reserved.
*/

/*
vidscreen V6. Testprogram for BScreen:custom modes: DVD and VCD overscanning modes.

written by Rudolf Cornelissen 3/2003-12/2003

Notes:
- The used timing is taken from BeTVOut, and is modifed to be compatible with the Matrox cards.
- While vertical timing for the secondary head is completely overwritten internal by the driver
  (except for v_display), the horizontal timing influences width and overscanning percentage of
  the TV output on G450/G550 for all modes, and for all but the modes used in this app on G400.
- The pixelclock set is only used on the primary head, and on the secondary head if it's _not_
  in TVout mode. In TVout mode the driver calculates the pixelclock itself. On G400 there is
  no guarantee that a valid pixelclock can be set for modes other than the modes used here.
- You may set a vertical resolution >= 480: TVout will clip at the max. amount of displayable 
  horizontal lines for the standard. It will include the overscanning range. (G400-G550)
- You may set a horizontal resolution of >=640 and <= 1022. If you go above 1022 the driver 
  will clip at 1022 pixels, including overscanning range. (G450/G550)
- On G400 a horizontal resolution of >=640 and <=1024 may be used, though no guarantees exist
  all possible modes will actually work (pixelclock constraints may not be met sometimes).
- On G400 the behaviour as described here applies as long as you keep the new flag TV_VIDEO set.
  If you clear it, the driver will set Desktop modes instead of Video modes:
  - Desktop modes underscan (black border around the screen) and have the flickerfilter enabled;
  - Video modes overscan (missing pieces of the picture around the screen) and have the
    flickerfilter disabled.
  In Desktop mode the timing for the secondary head is completely overwritten internal by the
  driver (except for h_display and v_display). It tries to correctly center and underscan the
  mode selected, though no guarantees can be given for other than the standard Desktop modes
  supported (being 640x480, 800x600 and 1024x768). 
- This program has been tested and written for G400, G450 and G550 cards using the V0.15 matrox
  driver.
*/

#include <stdio.h>
#include <string.h>

#include <be/game/WindowScreen.h>
#include <be/app/Application.h>
#include <be/interface/Screen.h>
#include <Bitmap.h>

#include "TVOut.h"
#include "Utility.h"

/* Matrox driver dualhead extensions to flags */
#define DUALHEAD_OFF (0<<6)
#define DUALHEAD_CLONE (1<<6)
#define DUALHEAD_ON (2<<6)
#define DUALHEAD_SWITCH (3<<6)
#define DUALHEAD_BITS (3<<6)
#define DUALHEAD_CAPABLE (1<<8)
#define TV_BITS (3<<9)
#define TV_MON (0<<9
#define TV_PAL (1<<9)
#define TV_NTSC (2<<9)
#define TV_CAPABLE (1<<11)
#define TV_VIDEO (1<<12)


class MatroxTVOut : public TVOut {
	public:
		MatroxTVOut(BMessage &settings);
		virtual ~MatroxTVOut();

		virtual status_t InitCheck();

		virtual status_t SetEnabled(bool enabled);
		virtual bool IsEnabled() const;

		virtual status_t SetMode(int32 mode);
		virtual status_t QueryModes(BMessage &modes);

		virtual void GetName(char *buffer, size_t bufferSize);
		virtual void GetInfo(char *buffer, size_t bufferSize);

	private:
		void UpdateTVOut();

		bool 			fIsEnabled;
		int32			fMode;
		display_mode	fOldMode;
		int32			fWorkspace;
};


MatroxTVOut::MatroxTVOut(BMessage &settings)
	: TVOut(settings),
	fIsEnabled(false)
	
{
	fMode = 1;
}


MatroxTVOut::~MatroxTVOut()
{
}


status_t 
MatroxTVOut::InitCheck()
{
}


void 
MatroxTVOut::UpdateTVOut()
{
	display_mode mode = fOldMode;
	BScreen screen;

	double refresh = (mode.timing.pixel_clock * 1000.0)
		/ (mode.timing.h_total * mode.timing.v_total);

	switch (fMode) {
		case 1:		// PAL 2x VCD (768 x 576)
			mode.virtual_width = 768;			//H_ACTIVE
			mode.timing.h_display = 768;
			mode.virtual_height = 576;			//V_ACTIVEI
			mode.timing.v_display = 576;
			mode.timing.h_sync_start = 848;		//set for centered TV output
			mode.timing.h_sync_end = 848+24;	//delta is BT H_BLANKI
			mode.timing.h_total = 944;			//BT H_CLKI
			mode.timing.v_sync_start = 579;		//set for centered sync pulse
			mode.timing.v_sync_end = 579+42;	//delta is BT V_BLANKI
			mode.timing.v_total = 625;			//BT V_LINESI
			mode.timing.pixel_clock = (uint32)((944 * 625 * refresh) / 1000); //refresh
			mode.flags &= ~(DUALHEAD_BITS | TV_BITS);
			mode.flags |= (TV_VIDEO | TV_PAL | DUALHEAD_CLONE);
			printf("\nset PAL 2x VCD mode (768 x 576)\n");
			screen.SetMode(&mode, false);
			break;

		case 2:		// NTSC 2x VCD (640 x 480)
			mode.virtual_width = 640;			//BT H_ACTIVE
			mode.timing.h_display = 640;
			mode.virtual_height = 480;			//BT  V_ACTIVEI
			mode.timing.v_display = 480;
			mode.timing.h_sync_start = 728;		//set for CH/BT compatible TV output
			mode.timing.h_sync_end = 728+24;	//delta is BT H_BLANKI
			mode.timing.h_total = 784;			//BT H_CLKI
			mode.timing.v_sync_start = 490;		//set for centered sync pulse
			mode.timing.v_sync_end = 490+25;	//delta is BT V_BLANKI
			mode.timing.v_total = 525;			//BT V_LINESI
			mode.timing.pixel_clock = (uint32)((784 * 525 * refresh) / 1000); //refresh
			mode.flags &= ~(DUALHEAD_BITS | TV_BITS);
			mode.flags |= (TV_VIDEO | TV_NTSC | DUALHEAD_CLONE);
			printf("\nset NTSC 2x VCD mode (640 x 480)\n");
			screen.SetMode(&mode, false);
			break;

		case 3:		// PAL DVD (720 x 576)
			mode.virtual_width = 720;			//BT H_ACTIVE
			mode.timing.h_display = 720;
			mode.virtual_height = 576;			//BT  V_ACTIVEI
			mode.timing.v_display = 576;
			mode.timing.h_sync_start = 736;		//set for centered sync pulse
			mode.timing.h_sync_end = 736+112;	//delta is BT H_BLANKI
			mode.timing.h_total = 888;			//BT H_CLKI
			mode.timing.v_sync_start = 579;		//set for centered sync pulse
			mode.timing.v_sync_end = 579+42;	//delta is BT V_BLANKI
			mode.timing.v_total = 625;			//BT V_LINESI
			mode.timing.pixel_clock = (uint32)((888 * 625 * refresh) / 1000); //refresh
			mode.flags &= ~(DUALHEAD_BITS | TV_BITS);
			mode.flags |= (TV_VIDEO | TV_PAL | DUALHEAD_CLONE);
			printf("\nset PAL DVD mode (720 x 576)\n");
			screen.SetMode(&mode, false);
			break;

		case 4:		// NTSC DVD (720 x 480)
			mode.virtual_width = 720;			//H_ACTIVE
			mode.timing.h_display = 720;
			mode.virtual_height = 480;			//V_ACTIVEI
			mode.timing.v_display = 480;
			mode.timing.h_sync_start = 736;		//do not change!
			mode.timing.h_sync_end = 736+112;	//delta is H_sync_pulse
			mode.timing.h_total = 888;			//BT H_TOTAL
			mode.timing.v_sync_start = 490;		//set for centered sync pulse
			mode.timing.v_sync_end = 490+20;	//delta is V_sync_pulse
			mode.timing.v_total = 525;			//CH V_TOTAL
			mode.timing.pixel_clock = (uint32)((888 * 525 * refresh) / 1000); //refresh
			mode.flags &= ~(DUALHEAD_BITS | TV_BITS);
			mode.flags |= (TV_VIDEO | TV_NTSC | DUALHEAD_CLONE);
			printf("\nset NTSC DVD mode (720 x 480)\n");
			screen.SetMode(&mode, false);
			break;
	}
}


status_t 
MatroxTVOut::SetEnabled(bool enabled)
{
	if (fIsEnabled == enabled)
		return B_OK;

	status_t status = B_OK;

	if (fIsEnabled) {
		// restore original display mode
		BScreen screen;
		screen.SetMode(fWorkspace, &fOldMode);
	} else {
		// save current display mode
		BScreen screen;
		screen.GetMode(&fOldMode);
		fWorkspace = current_workspace();

		UpdateTVOut();
	}

	if (status == B_OK)
		fIsEnabled = enabled;

	return status;
}


bool 
MatroxTVOut::IsEnabled() const
{
	return fIsEnabled;
}


status_t 
MatroxTVOut::SetMode(int32 mode)
{
	if (fMode == mode)
		return B_OK;

	fMode = mode;
	if (fIsEnabled)
		UpdateTVOut();

	return B_OK;
}


status_t 
MatroxTVOut::QueryModes(BMessage &modes)
{
	const struct { const char *name; int32 mode; } resolutions[] = {
		{"640x480 NTSC VCD/MPEG1", 2},
		{"720x480 NTSC DVD/MPEG2", 4},
		{"768x576 PAL VCD/MPEG1", 1},
		{"720x576 PAL DVD/MPEG2", 3},
		{NULL, -1}
	};

	for (int32 i = 0; resolutions[i].name != NULL; i++) {
		modes.AddString("name", resolutions[i].name);
		modes.AddInt32("mode", resolutions[i].mode);
	}
	modes.AddInt32("ntsc_modes", 2);
	modes.AddInt32("current", fMode);

	return B_OK;
}


void 
MatroxTVOut::GetName(char *buffer, size_t bufferSize)
{
	// R5 has no strlcpy()
	strncpy(buffer, "Matrox", bufferSize - 1);
	buffer[bufferSize - 1] = '\0';
}


void 
MatroxTVOut::GetInfo(char *buffer, size_t bufferSize)
{
	accelerant_device_info info;
	BScreen screen;
	if (screen.GetDeviceInfo(&info) != B_OK)
		strcpy(info.name, "written by Rudolf Cornelissen");

	// R5 has no strlcpy()
	strncpy(buffer, info.name, bufferSize - 1);
	buffer[bufferSize - 1] = '\0';
}


//	#pragma mark -


TVOut *
GetMatroxTVOut(BMessage &settings)
{
	// check for Matrox card and Rudolf's drivers!

	if (check_for_image(1, "mga.driver") != B_OK)
		return NULL;

	return new MatroxTVOut(settings);
}

