/* nVidia - TVOut class for nVidia cards using the riva_iic driver.
**		Based on code originally developed by Rudolf Cornelissen and
**		Michael Pfeiffer for BeTVOut.
**
** Copyright (c) 2004 pinc Software. All Rights Reserved.
*/


#include <Application.h>
#include <Window.h>
#include <View.h>
#include <Bitmap.h>
#include <PopUpMenu.h>
#include <MenuItem.h>
#include <Deskbar.h>
#include <Roster.h>
#include <Screen.h>
#include <Alert.h>
#include <Resources.h>

#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include "TVOut.h"
#include "riva_iic.h"
#include "BT868_9.h"
#include "Chrontel.h"
#include "Transfer.h"
#include "Utility.h"


const uint16 kBeTVOutVersion = 120;


class Chip {
	public:
		Chip(int device, riva_data_t &rivaData);
		virtual ~Chip();

		virtual status_t InitCheck() = 0;

		virtual status_t EnableTVOut(bool saveMode = true) = 0;
		virtual status_t DisableTVOut() = 0;		

		virtual status_t SetMode(int32 mode);
		virtual status_t QueryModes(BMessage &modes);

		virtual void GetInfo(char *buffer, size_t bufferSize) = 0;

		int32 Mode() const { return fMode; }
		void SetDevice(int device);

		static Chip *GetChip(int device, riva_data_t &rivaData);

	protected:
		int			fDevice;
		riva_data_t	&fRivaData;
		int			fMode;
		bool		fIsEnabled;
};


class BrookTreeChip : public Chip {
	public:
		BrookTreeChip(int device, riva_data_t &rivaData, int type, int version);
		virtual ~BrookTreeChip();

		virtual status_t InitCheck();

		virtual status_t EnableTVOut(bool saveMode = true);
		virtual status_t DisableTVOut();		

		virtual status_t SetMode(int32 mode);

		virtual void GetInfo(char *buffer, size_t bufferSize);

	private:
		status_t UpdateTVOut();

		int				fType, fVersion;
		display_mode	fOldMode;
		int32			fWorkspace;
		bool			fNTSC;
};


class nVidiaTVOut : public TVOut {
	public:
		nVidiaTVOut(BMessage &settings);
		virtual ~nVidiaTVOut();

		virtual status_t InitCheck();

		virtual status_t SetEnabled(bool enabled);
		virtual bool IsEnabled() const;

		virtual status_t SetMode(int32 mode);
		virtual status_t QueryModes(BMessage &modes);
		virtual void WorkspaceChanged();

		virtual void GetName(char *buffer, size_t bufferSize);
		virtual void GetInfo(char *buffer, size_t bufferSize);

	private:
		bool OpenDevice();
		status_t EnableTVOut(bool workspaceChanged = false);
		status_t DisableTVOut(bool workspaceChanged = false);

		int				fDevice;
		riva_data_t		fRivaData;
		Chip			*fChip;
		bool 			fIsEnabled;
		int32			fWorkspace;
		int32			fMode;
		const char		*fCardName;
};


//-------------------------------


int
nonVESAsetVESA640(BScreen* screenptr, display_mode new_mode)
{
	//set VESA640: (hex values are exact copies of direct register programmings)
	new_mode.virtual_width = 640;
	new_mode.virtual_height = 480;
	new_mode.timing.h_display = 640;
	new_mode.timing.v_display = 480;
	
	new_mode.timing.h_sync_start = (0x53 * 8);//byte oriented!
	new_mode.timing.h_sync_end = ((0x1f + (0x53 && 0xe0)) * 8);//byte oriented!
	new_mode.timing.h_total = (0x5F * 8);//byte oriented! (val. *must* be set!)
	new_mode.timing.v_sync_start = 0x1ea;
	new_mode.timing.v_sync_end = ((0x1ea && 0xff0) + 0x00c);
	new_mode.timing.v_total = 0x20b;
	new_mode.timing.pixel_clock = ((((0x5F * 8) * 0x20b)* 60)/1000);// in kHz!

	return (*screenptr).SetMode(&new_mode, false);
}


int
nonVESAsetVESA800 (BScreen* screenptr, display_mode new_mode)
{
	//set VESA800: (hex values are exact copies of direct register programmings)
	new_mode.virtual_width = 800;
	new_mode.virtual_height = 600;
	new_mode.timing.h_display = 800;
	new_mode.timing.v_display = 600;
	
	new_mode.timing.h_sync_start = (0x6a * 8);//byte oriented!
	new_mode.timing.h_sync_end = ((0x1a + (0x6a && 0xe0)) * 8);//byte oriented!
	new_mode.timing.h_total = (0x7f * 8);//byte oriented! (this value *must* be set!)
	new_mode.timing.v_sync_start = 0x259;
	new_mode.timing.v_sync_end = ((0x259 && 0xff0) + 0x00d);
	new_mode.timing.v_total = 0x272;
	new_mode.timing.pixel_clock = (((((0x7F) * 8) * 0x272)* 60)/1000);// in kHz!

	return (*screenptr).SetMode(&new_mode, false);
}


int
checkVESAmode (int fd, BScreen* screenptr, display_mode new_mode,
	display_mode org_mode, riva_data_t *ddata)
{
	int status = B_OK;
	riva_data_t tmpddata = *ddata;

	//assume we are accessing the primary VGA card:
	ddata->devID = 0;
	//get current RIVA display mode:
	ioctl(fd, rivaGetMode,ddata);
	//(current Be Desktop card resolution already fetched..)

	//do prim/sec test:
	if ((ddata->displ_mode.displ_timing.h_display != new_mode.virtual_width) |
		(ddata->displ_mode.displ_timing.v_display != new_mode.virtual_height))
	{
		//RIVA in different resolution than Be Desktop:
		ddata->devID = 1; //also means status of RIVA card is known as of now..
	}

	//do destructive VESAmode check:
	if (new_mode.virtual_width == 640)
	{
		//Be Desktop (primary card) is at 640x480 mode...
		status = nonVESAsetVESA800 (screenptr, new_mode);
	}
	else
	{
		//Be Desktop (primary card) is NOT at 640x480 mode...
		status = nonVESAsetVESA640 (screenptr, new_mode);
	}

	//get newest (temporary) Be Desktop card resolution:
	(*screenptr).GetMode(&new_mode);
	//get newest (temporary) RIVA display mode:
	ioctl(fd, rivaGetMode,&tmpddata);

	//do prim/sec test again:
	if ((tmpddata.displ_mode.displ_timing.h_display != new_mode.virtual_width) |
		(tmpddata.displ_mode.displ_timing.v_display != new_mode.virtual_height))
	{
		//RIVA in different resolution than Be Desktop:
		ddata->devID = 1; //also means status of RIVA card is known as of now..
	}

	//restore original settings:
	(*screenptr).SetMode(&org_mode, false);
	
	if (status == B_OK) return (0);	//desktop is in driver (non-VESA) mode
	else return (1); //desktop is in VESA mode..
}


//	#pragma mark -


Chip::Chip(int device, riva_data_t &rivaData)
	:
	fDevice(device),
	fRivaData(rivaData),
	fIsEnabled(false)
{
}


Chip::~Chip()
{
}


status_t 
Chip::SetMode(int32 mode)
{
	fMode = mode;
}


status_t 
Chip::QueryModes(BMessage &modes)
{
}


void 
Chip::SetDevice(int device)
{
	fDevice = device;
}


Chip *
Chip::GetChip(int device, riva_data_t &rivaData)
{
	// unlock RIVA extended VGA registers...
	ioctl(device, rivaUnlock);

	// check for one or two available (wired) IIC busses.
	// On RIVA TNT and later, two could be there...
	int RIVAexists = 0;
	int bus0wired = 0, bus0sane = 0, bus1wired = 0, bus1sane = 0;

	printf("RIVA IICbus 0: ");
	rivaData.iic.busnr = 0;
	if (!ioctl(device, rivaChkBus, &rivaData)) {
		printf("detected.\n");
		RIVAexists = 1;
		bus0wired = 1;
	} else {
		printf("not detected.\n");
	}

	printf("RIVA IICbus 1: ");
	rivaData.iic.busnr = 1;
	if (!ioctl(device, rivaChkBus, &rivaData)) {
		printf("detected.\n");
		RIVAexists = 1;
		bus1wired = 1;
	} else {
		printf("not detected.\n");
	}

	if (!RIVAexists) {
		printf("No RIVA IIC bus found, exiting...\n");
		return NULL;
	}

	// find at least one sane IIC bus... (test this by sending an unused adress...)
	if (!FindSaneBus(device, bus0wired, bus1wired, &bus0sane, &bus1sane, &rivaData)) {
		printf("Unexpected results, exiting...\n");
		return NULL;
	}

	//now find a TV-out chip...
	int type, version;
	int	chfound = 0;
	byte chbus = 0, chaddr = 0, chversion;

	//check for BT or CX chip:
	int btfound = FindBT(device, &rivaData);
	if (btfound) {
		//read chiptype and version number:
		int input;
		if ((input = ReadType(device, &version, &type, &rivaData))) {
			//something went wrong:
			if (input == 1) {
				// IIC error encountered:
				return NULL;
			} else {
				// The chip apparantly is not a TVout chip:
				btfound = 0;
			}
		} else {
			// no errors found:
			printf("ChipID = %d, chip version = %d: Chiptype is ", type, version);
			switch (type) {
				case 0: printf("bt868.\n");
						break;
				case 1: printf("bt869.\n");
						break;
				case 2: printf("cx25870.\n");
						break;
				case 3: printf("cx25871.\n");
						break;
				default:printf("unknown, exiting...\n");
						return NULL;
			}
		}
		return new BrookTreeChip(device, rivaData, type, version);
	}

#if 0
	// check for CH chip:
	if (!btfound) {
		if (bus0sane) {
			ddata->iic.busnr = 0;
			chbus = 0;
			chfound = ch_detect(fDevice, chbus, &chaddr, &chversion, &fRivaData);
		}
		if (bus1sane && !chfound)
		{
			ddata->iic.busnr = 1;
			chbus = 1;
			chfound = ch_detect(fd, chbus, &chaddr, &chversion, ddata);
		}

		if (chfound)
		{
			printf("\nFound one on bus %d, adress $%x;\n",chbus,chaddr);
			printf("Chiptype is ch%s.\n",ch_model_name(chversion));
		}
		else //no TVout chip at all...
		{
			printf("\nNo supported TVout chip found, exiting...\n");
			close (fd);
			return(B_OK);
		}
	}
#endif

	return NULL;
}


//	#pragma mark -


BrookTreeChip::BrookTreeChip(int device, riva_data_t &rivaData, int type, int version)
	: Chip(device, rivaData),
	fType(type),
	fVersion(version)
{
	fMode = 12;
	fNTSC = false;
}


BrookTreeChip::~BrookTreeChip()
{
}


status_t 
BrookTreeChip::InitCheck()
{
}


status_t
BrookTreeChip::UpdateTVOut()
{
	BScreen screen;
	display_mode new_mode = fOldMode;
	int monstat;

printf("bt_update, enabled %d!\n", fIsEnabled);

	// fixme: reset BT should be here...
	// (but: beware of the 'locked SDA' syndrome then!)
	if (KillClock_Blackout_BT(fDevice, fType, &fRivaData))
		return B_ERROR;

	// if TV output signals is set to 'autodetect' then do it:
//		if (OutMethod == 3)
	{
		monstat = ReadBT_ConnectStat(fDevice, &fRivaData);
		if (monstat == -1)
			return B_ERROR;
	}

	// 'slowdown RIVA clock' if TVout is requested:
	// makes sure TVout chip is not being overclocked! (VESA640/800 is safe..)
	// NOTE: it looks like this is unnesessary after all..
	switch (fMode) {
		case 2:
		case 5:
		case 11:
			nonVESAsetVESA640(&screen, fOldMode);
			break;
		case 3:
		case 6:
		case 9:
		case 10:
		case 12:
			nonVESAsetVESA800(&screen, fOldMode);
			break;
	}

	//Manual config is non-dependent of the state of the PAL hardware input pin;
	//Also SDA lockups occur when setting EN_XCLK after autoconfig!
	//Make sure PAL_MD=0 for NTSC and PAL_MD = 1 for PAL...
	switch (fMode) {
		case 1:
		case 2:
			if (InitBT_NTSC640(fDevice, &fRivaData))
				return B_ERROR;
			break;
		case 3:
			if (InitBT_NTSC800(fDevice, &fRivaData))
				return B_ERROR;
			break;
		case 4:
			if (InitBT_PAL800(fDevice, &fRivaData))
				return B_ERROR;
			break;
		case 5:
			if (InitBT_PAL640(fDevice, &fRivaData))
				return B_ERROR;
			break;
		case 6:
			if (InitBT_PAL800(fDevice, &fRivaData))
				return B_ERROR;
			break;
		case 9:
			if (InitBT_NTSC720(fDevice, &fRivaData))
				return B_ERROR;
			break;
		case 10:
			if (InitBT_PAL720(fDevice, &fRivaData))
				return B_ERROR;
			break;
		case 11:
			if (InitBT_NTSC640_OS(fDevice, &fRivaData))
				return B_ERROR;
			break;
		case 12:
			if (InitBT_PAL800_OS(fDevice, &fRivaData))
				return B_ERROR;
			break;
	}	

	// modify BT Hphase signal to center TV image...
	if (SetBT_Hphase(fDevice, fMode, &fRivaData))
		return B_ERROR;

	// set Macro mode 0...
	switch (fMode) {
		case 1:
		case 2:
		case 3:
		case 9:
		case 11:
			fNTSC = true;
			if (SetMacro(fDevice, &fRivaData, 0, 0) != 0)	// NTSC
				return B_ERROR;
			break;
		case 4:
		case 5:
		case 6:
		case 10:
		case 12:
			fNTSC = false;
			if (SetMacro(fDevice, &fRivaData, 1, 0) != 0)	// PAL
				return B_ERROR;
			break;
	}

	// setup BT for TVout (filters, DAC_muxes...)
	if (SetupBTforTVout(fDevice, 3/*OutMethod*/, monstat, 1/*ffilter*/, &fRivaData))
		return B_ERROR;

   	printf("\nOption sent...");

	switch (fMode) {	//both BT and CH get here..
		case 1:
			if (DoTestImage(fDevice, &fRivaData))
				return B_ERROR;
		   	printf(" and NTSC testimage activated.\n");
			break;
		case 2: //reprogram RIVA for NTSC 640x480 @ 60Hz mode
				ioctl(fDevice, rivaNtsc640, &fRivaData);
				printf(" and NTSC 640x480 @ 60Hz mode activated.\n");
				break;
		case 3: //reprogram RIVA for NTSC 800x600 @ 60Hz mode
				ioctl(fDevice, rivaNtsc800, &fRivaData);
				printf(" and NTSC 800x600 @ 60Hz mode activated.\n");
				break;
		case 4:
			if (DoTestImage(fDevice, &fRivaData))
				return B_ERROR;
	   		printf(" and PAL testimage activated.\n"); 
			break;
		case 5: //reprogram RIVA for PAL 640x480 @ 50Hz mode
				ioctl(fDevice, rivaPal640, &fRivaData);
				printf(" and PAL 640x480 @ 50Hz mode activated.\n");
				break;
		case 6: //reprogram RIVA for PAL 800x600 @ 50Hz mode
				ioctl(fDevice, rivaPal800, &fRivaData);
				printf(" and PAL 800x600 @ 50Hz mode activated.\n");
				break;
		case 9:
			//reprogram RIVA for NTSC 720x480 @ 60Hz mode:
			//(tested on TNT2 with BT869)
			new_mode.virtual_width = 720;			//H_ACTIVE
			new_mode.timing.h_display = 720;
			new_mode.virtual_height = 480;			//V_ACTIVEI
			new_mode.timing.v_display = 480;

			new_mode.timing.h_sync_start = 744;		//do not change!
			new_mode.timing.h_sync_end = 744+144;	//delta is H_sync_pulse
			new_mode.timing.h_total = 888;			//BT H_TOTAL
			new_mode.timing.v_sync_start = 490;		//set for centered sync pulse
			new_mode.timing.v_sync_end = 490+26;	//delta is V_sync_pulse
			new_mode.timing.v_total = 525;			//CH V_TOTAL
			new_mode.timing.pixel_clock = ((888 * 525 * 60) / 1000); //refresh

			screen.SetMode(&new_mode, false);
			//switch RIVA to slave-mode for TVout chip:
			ioctl(fDevice, rivaEnableTVOut);
			printf(" and NTSC 720x480 @ 60Hz overscanning mode activated.\n");
			break;
		case 10:
			//reprogram RIVA for PAL 720x576 @ 50Hz mode:(a bit modified for CH)
				new_mode.virtual_width = 720;			//BT H_ACTIVE
				new_mode.timing.h_display = 720;
				new_mode.virtual_height = 576;			//BT  V_ACTIVEI
				new_mode.timing.v_display = 576;
				new_mode.timing.h_sync_start = 744;		//set for centered sync pulse
				new_mode.timing.h_sync_end = 744+140;	//delta is BT H_BLANKI
				new_mode.timing.h_total = 888;			//BT H_CLKI
				new_mode.timing.v_sync_start = 579;		//set for centered sync pulse
				new_mode.timing.v_sync_end = 579+42;	//delta is BT V_BLANKI
				new_mode.timing.v_total = 625;			//BT V_LINESI
				new_mode.timing.pixel_clock = ((888 * 625 * 50) / 1000); //refresh
				screen.SetMode(&new_mode, false);
				//switch RIVA to slave-mode for TVout chip:
				ioctl(fDevice, rivaEnableTVOut);
				printf(" and PAL 720x576 @ 50Hz overscanning mode activated.\n");
				break;
		case 11://reprogram RIVA for NTSC 640x480 (OverScanning) @ 60Hz mode:
				new_mode.virtual_width = 640;			//BT H_ACTIVE
				new_mode.timing.h_display = 640;
				new_mode.virtual_height = 480;			//BT  V_ACTIVEI
				new_mode.timing.v_display = 480;
				new_mode.timing.h_sync_start = 744;		//set for CH/BT compatible TV output
				new_mode.timing.h_sync_end = 744+20;	//delta is BT H_BLANKI
				new_mode.timing.h_total = 784;			//BT H_CLKI
				new_mode.timing.v_sync_start = 490;		//set for centered sync pulse
				new_mode.timing.v_sync_end = 490+25;	//delta is BT V_BLANKI
				new_mode.timing.v_total = 525;			//BT V_LINESI
				new_mode.timing.pixel_clock = ((784 * 525 * 60) / 1000); //refresh
				screen.SetMode(&new_mode, false);
				//switch RIVA to slave-mode for TVout chip:
				ioctl(fDevice, rivaEnableTVOut);
				printf(" and NTSC 640x480 @ 60Hz overscanning mode activated.\n");
				break;
		case 12://reprogram RIVA for PAL 768x576 @ 50Hz mode:
				new_mode.virtual_width = 768;			//H_ACTIVE
				new_mode.timing.h_display = 768;
				new_mode.virtual_height = 576;			//V_ACTIVEI
				new_mode.timing.v_display = 576;
				new_mode.timing.h_sync_start = 848;		//set for centered TV output
				new_mode.timing.h_sync_end = 848+20;	//delta is BT H_BLANKI
				new_mode.timing.h_total = 944;			//BT H_CLKI
				new_mode.timing.v_sync_start = 579;		//set for centered sync pulse
				new_mode.timing.v_sync_end = 579+42;	//delta is BT V_BLANKI
				new_mode.timing.v_total = 625;			//BT V_LINESI
				new_mode.timing.pixel_clock = ((944 * 625 * 50) / 1000); //refresh
				screen.SetMode(&new_mode, false);
				//switch RIVA to slave-mode for TVout chip:
				ioctl(fDevice, rivaEnableTVOut);
				printf(" and PAL 768x576 @ 50Hz overscanning mode activated.\n");
				break;
	}

	switch (fMode) {
		case 2:
		case 3:
		case 5:
		case 6:
		case 9:
		case 10:
		case 11:
		case 12://select 'singlehead' or 'dualhead clone' mode:
/*				if (dualclone)
				{
					//turn on VGA:
					ioctl(fd, rivaVgaOn);
				}
				else
				{
*/					//turn off VGA:
					ioctl(fDevice, rivaVgaOff);
//				}
				break;
/*		case 7: //select 'VGA only' mode:
				if (VGAonlyMode (fd, Type, btfound, chbus, chaddr, screenptr,
																 org_mode, ddata))
				{
					close (fd);
					return (B_OK);
				}
				break;*/
	}
}


status_t
BrookTreeChip::EnableTVOut(bool saveMode)
{
	BScreen screen;
	if (saveMode) {
		screen.GetMode(&fOldMode);
		fWorkspace = current_workspace();
	}

	UpdateTVOut();
	fIsEnabled = true;
	return B_OK;
}


status_t
BrookTreeChip::DisableTVOut()
{
	// prevent BT from being overclocked by VGA-only modes & black-out TV-out..
	if (KillClock_Blackout_BT(fDevice, fType, &fRivaData)) 
		puts("bt killclock blackout failed");

	// turn on VGA only:
	ioctl(fDevice, rivaVgaOnly);

	BScreen screen;
	screen.SetMode(fWorkspace, &fOldMode, false);

	// set PAL or NTSC test image when TV out is disabled
	if (fNTSC)
		InitBT_NTSC640(fDevice, &fRivaData);
	else
		InitBT_PAL800(fDevice, &fRivaData);

	DoTestImage(fDevice, &fRivaData);

	fIsEnabled = false;
	return B_OK;
}


status_t 
BrookTreeChip::SetMode(int32 mode)
{
	if (fMode == mode)
		return B_OK;

	Chip::SetMode(mode);
	fNTSC = mode < 5 || mode == 9 || mode == 11;

	if (fIsEnabled)
		UpdateTVOut();
}


void 
BrookTreeChip::GetInfo(char *buffer, size_t bufferSize)
{
	snprintf(buffer, bufferSize, "BrookTree 868/9 Type %d Version 0x%x", fType, fVersion);
}


//	#pragma mark -


nVidiaTVOut::nVidiaTVOut(BMessage &settings)
	: TVOut(settings),
	fDevice(-1),
	fChip(NULL),
	fIsEnabled(false),
	fWorkspace(-1),
	fCardName("Graphics Adapter")
{
	fMode = 12;

	int32 mode;
	if (settings.FindInt32("nvidia_mode", &mode) == B_OK)
		fMode = mode;
}


nVidiaTVOut::~nVidiaTVOut()
{
	delete fChip;
	close(fDevice);
}


status_t 
nVidiaTVOut::InitCheck()
{
	return B_OK;
}


status_t 
nVidiaTVOut::SetEnabled(bool enabled)
{
	if (fIsEnabled == enabled)
		return B_OK;

	status_t status;

	if (fIsEnabled)
		status = DisableTVOut();
	else
		status = EnableTVOut();

	if (status == B_OK)
		fIsEnabled = enabled;

	return status;
}


bool 
nVidiaTVOut::IsEnabled() const
{
	return fIsEnabled;
}


status_t
nVidiaTVOut::SetMode(int32 mode)
{
	if (mode < 1 || mode > 12)
		return B_BAD_VALUE;

	fMode = mode;
	if (fChip != NULL)
		return fChip->SetMode(mode);

	return B_OK;
}


status_t 
nVidiaTVOut::QueryModes(BMessage &modes)
{
	const struct { const char *name; int32 mode; } resolutions[] = {
		{"640x480 NTSC", 2},
		{"800x600 NTSC", 3},
		{"720x480 NTSC DVD/MPEG2", 9},
		{"640x480 NTSC VCD/MPEG1", 11},
		{"640x480 PAL", 5},
		{"800x600 PAL", 6},
		{"720x576 PAL DVD/MPEG2", 10},
		{"768x576 PAL VCD/MPEG1", 12},
		{NULL, -1}
	};

	for (int32 i = 0; resolutions[i].name != NULL; i++) {
		modes.AddString("name", resolutions[i].name);
		modes.AddInt32("mode", resolutions[i].mode);
	}
	modes.AddInt32("ntsc_modes", 4);
	modes.AddInt32("current", fChip != NULL ? fChip->Mode() : fMode);

	return B_OK;
}


bool 
nVidiaTVOut::OpenDevice()
{
	if (fDevice >= 0)
		return true;

	fDevice = open("/dev/video/riva_iic/1", O_RDWR);
	if (fDevice < B_OK) {
		BAlert *alert = new BAlert("TVOut", "Could not open riva_iic device.", "Ok");
		alert->Go(NULL);
		return false;
	}

	// setup 'magic value' to unlock driver: 
	// (to keep compatibility & prevent crashes with older versions of driver and app...)
	fRivaData.drv_version = 0xff;	
	fRivaData.app_version = kBeTVOutVersion;

	// check version of driver...
	if (ioctl(fDevice, rivaVersion, &fRivaData) == 0) {
		if (fRivaData.drv_version != 0xff && fRivaData.drv_version >= 23) {
			// we have the correct version, get the chip
			fChip = Chip::GetChip(fDevice, fRivaData);
			if (fChip != NULL) {
				fChip->SetMode(fMode);

				switch (fRivaData.dev.device_id) {
					case 0x0020:
						fCardName = "Riva TNT1";
						fRivaData.type = 1;
						break;
					case 0x0028:
						fCardName = "Riva TNT2";
						fRivaData.type = 2;
						break;
					case 0x002d:
						fCardName = "Riva TNT2-M64";
						fRivaData.type = 3;
						break;
					case 0x0101:
						fCardName = "GeForce 256";
						fRivaData.type = 1;
						break;
					case 0x0110:
						fCardName = "GeForce 2 MX (200/400)";
						fRivaData.type = 1;
						break;
					default:
						fRivaData.type = 1;
						break;
				}

				return true;
			}

			BAlert *alert = new BAlert("nvtv", "Did not find supported TV-Out chip.", "Ok");
			alert->Go(NULL);

			close(fDevice);
			fDevice = -1;
			return false;
		}
	}

	BAlert *alert = new BAlert("nvtv", "riva_iic has wrong version (0.23 or newer needed).", "Ok");
	alert->Go(NULL);

	close(fDevice);
	fDevice = -1;
	return false;
}


status_t
nVidiaTVOut::EnableTVOut(bool workspaceChanged)
{
	if (!OpenDevice())
		return B_ERROR;

	if (!workspaceChanged)
		fWorkspace = current_workspace();

	return fChip->EnableTVOut(!workspaceChanged);
}


status_t
nVidiaTVOut::DisableTVOut(bool workspaceChanged)
{
	if (fChip == NULL)
		return B_NO_INIT;

	status_t status = fChip->DisableTVOut();

	if (!workspaceChanged)
		fWorkspace = -1;

	return status;
}


void 
nVidiaTVOut::WorkspaceChanged()
{
	if (!fIsEnabled)
		return;

	if (current_workspace() == fWorkspace)
		EnableTVOut(true);
	else
		DisableTVOut(true);
}


void 
nVidiaTVOut::GetName(char *buffer, size_t bufferSize)
{
	OpenDevice();
		// this will sniff the chip data

	snprintf(buffer, bufferSize, "nVidia %s", fCardName);
}


void 
nVidiaTVOut::GetInfo(char *buffer, size_t bufferSize)
{
	if (fChip != NULL) {
		fChip->GetInfo(buffer, bufferSize);
	} else {
		// R5 has no strlcpy()
		strncpy(buffer, "Unsupported TV-Out Chip", bufferSize - 1);
		buffer[bufferSize - 1] = '\0';
	}
}


//	#pragma mark -


TVOut *
GetNVidiaTVOut(BMessage &settings)
{
	// check for known nvidia drivers (by name, we're lame)

	if (check_for_image(1, "nvidia") != B_OK
		&& check_for_image(1, "rivatnt") != B_OK
		&& check_for_image(1, "riva128") != B_OK
		&& check_for_image(1, "nv.driver") != B_OK)
		return NULL;

	return new nVidiaTVOut(settings);
}

